@@ -1,8 +1,5 @@
 /*
- * ossp-padsp - ossp DSP slave which forwards to pulseaduio
- *
- * Copyright (C) 2008-2010  SUSE Linux Products GmbH
- * Copyright (C) 2008-2010  Tejun Heo <tj@kernel.org>
+ * ossp-padsp - ossp DSP slave which forwards to PulseAudio
  *
  * This file is released under the GPLv2.
  */
@@ -22,6 +19,8 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <linux/limits.h>
+#include <stdlib.h>
 
 #include <pulse/pulseaudio.h>
 #include <sys/soundcard.h>
@@ -784,6 +783,58 @@ static ssize_t padsp_open(enum ossp_opcode opcode,
 	return 0;
 }
 
+static void notify_mmap_fill_play(size_t mmap_size, size_t pos, size_t bytes)
+{
+	while (bytes) {
+		ssize_t ret;
+		size_t count = min(bytes, mmap_size - pos);
+		struct ossp_notify event = { .magic = OSSP_NOTIFY_MAGIC,
+					     .opcode = OSSP_NOTIFY_FILL };
+
+		ossp_mmap_transfer[PLAY].pos = pos;
+		ossp_mmap_transfer[PLAY].bytes = count;
+
+		ret = write(ossp_notify_fd, &event, sizeof(event));
+		if (ret != sizeof(event)) {
+			if (errno != EPIPE)
+				err_e(-errno, "write to notify_fd failed");
+
+			padsp_done();
+			return;
+		}
+		sem_wait(&ossp_mmap_transfer[PLAY].sem);
+
+		bytes -= count;
+		pos = (pos + count) % mmap_size;
+	}
+}
+
+static void notify_mmap_store_rec(size_t mmap_size, size_t pos, size_t bytes)
+{
+	while (bytes) {
+		ssize_t ret;
+		size_t count = min(bytes, mmap_size - pos);
+		struct ossp_notify event = { .magic = OSSP_NOTIFY_MAGIC,
+					     .opcode = OSSP_NOTIFY_STORE };
+
+		ossp_mmap_transfer[REC].pos = pos;
+		ossp_mmap_transfer[REC].bytes = count;
+
+		ret = write(ossp_notify_fd, &event, sizeof(event));
+		if (ret != sizeof(event)) {
+			if (errno != EPIPE)
+				err_e(-errno, "write to notify_fd failed");
+
+			padsp_done();
+			return;
+		}
+		sem_wait(&ossp_mmap_transfer[REC].sem);
+
+		bytes -= count;
+		pos = (pos + count) % mmap_size;
+	}
+}
+
 static void mmap_fill_pstg(void)
 {
 	struct ring_buf *stg = &mmap_stg[PLAY];
@@ -812,6 +863,8 @@ static void mmap_fill_pstg(void)
 		bytes = space;
 	}
 
+	notify_mmap_fill_play(mmap_size, mmap_idx[PLAY] % mmap_size, bytes);
+
 	ring_manual_init(&mmap, mmap_map[PLAY], mmap_size,
 			 new_idx % mmap_size, bytes);
 
@@ -898,6 +951,8 @@ static void mmap_consume_rstg(void)
 		space -= todo;
 	}
 
+	notify_mmap_store_rec(mmap_size, fill_idx % mmap_size, ring_bytes(&mmap));
+
  skip:
 	mmap_idx[REC] = new_idx;
 }
@@ -972,16 +1027,13 @@ static void do_mmap_read(size_t bytes)
 
 static void stream_rw_callback(pa_stream *s, size_t length, void *userdata)
 {
-	int dir;
 	size_t size;
 
 	if (s == stream[PLAY]) {
-		dir = PLAY;
 		size = pa_stream_writable_size(s);
 		if (mmap_map[PLAY])
 			do_mmap_write(size);
 	} else if (s == stream[REC]) {
-		dir = REC;
 		size = pa_stream_readable_size(s);
 		if (mmap_map[REC])
 			do_mmap_read(size);
@@ -1482,7 +1534,7 @@ int main(int argc, char **argv)
 {
 	int rc;
 
-	ossp_slave_init(argc, argv);
+	ossp_slave_init("ossp-padsp", argc, argv);
 
 	page_size = sysconf(_SC_PAGE_SIZE);
 
