@@ -1,13 +1,10 @@
 /*
  * osspd - OSS Proxy Daemon: emulate OSS device using CUSE
  *
- * Copyright (C) 2008-2010  SUSE Linux Products GmbH
- * Copyright (C) 2008-2010  Tejun Heo <tj@kernel.org>
- *
  * This file is released under the GPLv2.
  */
 
-#define FUSE_USE_VERSION 28
+#define FUSE_USE_VERSION 35
 #define _GNU_SOURCE
 
 #include <assert.h>
@@ -22,6 +19,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
+#include <sys/ioctl.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
 #include <sys/epoll.h>
@@ -34,13 +32,6 @@
 #include "ossp.h"
 #include "ossp-util.h"
 
-/*
- * MMAP support needs to be updated to the new fuse MMAP API.  Disable
- * it for the time being.
- */
-#warning mmap support disabled for now
-/* #define OSSP_MMAP */
-
 #define DFL_MIXER_NAME		"mixer"
 #define DFL_DSP_NAME		"dsp"
 #define DFL_ADSP_NAME		"adsp"
@@ -127,8 +118,11 @@ struct ossp_stream {
 	int			vol[2][2];
 	int			vol_set[2][2];
 
-	off_t			mmap_off;
+	int			mmap_fd;
 	size_t			mmap_size;
+	void			*mmap;
+	void			*mmap_addr[2];
+	struct ossp_transfer	*mmap_transfer;
 
 	struct ossp_uid_cnt	*ucnt;
 	struct fuse_session	*se;	/* associated fuse session */
@@ -469,15 +463,6 @@ static int ioctl_prep_uarg(fuse_req_t req, void *in, size_t in_sz, void *out,
 		return;							\
 } while (0)
 
-#define IOCTL_RETURN(result, outp) do {					\
-	if ((outp) != NULL)						\
-		fuse_reply_ioctl(req, result, (outp), sizeof(*(outp)));	\
-	else								\
-		fuse_reply_ioctl(req, result, NULL, 0);			\
-	return;								\
-} while (0)
-
-
 /***************************************************************************
  * Mixer implementation
  */
@@ -709,7 +694,8 @@ static void mixer_simple_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
 		strncpy(info.id, id, sizeof(info.id) - 1);
 		strncpy(info.name, name, sizeof(info.name) - 1);
 		info.modify_counter = mixer->modify_counter;
-		IOCTL_RETURN(0, &info);
+		fuse_reply_ioctl(req, 0, &info, sizeof(info));
+		break;
 	}
 
 	case SOUND_OLD_MIXER_INFO: {
@@ -718,7 +704,8 @@ static void mixer_simple_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
 		PREP_UARG(NULL, &info);
 		strncpy(info.id, id, sizeof(info.id) - 1);
 		strncpy(info.name, name, sizeof(info.name) - 1);
-		IOCTL_RETURN(0, &info);
+		fuse_reply_ioctl(req, 0, &info, sizeof(info));
+		break;
 	}
 
 	case OSS_GETVERSION:
@@ -737,16 +724,16 @@ static void mixer_simple_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
 		goto puti;
 	puti:
 		PREP_UARG(NULL, &i);
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SOUND_MIXER_WRITE_RECSRC:
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	default:
 		*not_minep = 1;
-		return;
 	}
-	assert(0);
 }
 
 static void mixer_do_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
@@ -787,7 +774,8 @@ static void mixer_do_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
 		break;
 	default:
 		i = 0;
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		return;
 	}
 
 	init_mixer_cmd(&mxcmd, mixer);
@@ -837,7 +825,12 @@ static void mixer_do_ioctl(fuse_req_t req, struct ossp_mixer *mixer,
 	finish_mixer_cmd(&mxcmd);
 	free(osa);
 
-	IOCTL_RETURN(0, out_bufsz ? &mxcmd.rvol : NULL);
+	if (out_bufsz)
+		fuse_reply_ioctl(req, 0, &mxcmd.rvol, sizeof(mxcmd.rvol));
+	else
+		fuse_reply_ioctl(req, 0, NULL, 0);
+
+	return;
 
 err:
 	fuse_reply_err(req, -rc);
@@ -898,6 +891,79 @@ static void mixer_release(fuse_req_t req, struct fuse_file_info *fi)
  * Stream implementation
  */
 
+static int os_create_shared_memory(struct ossp_stream *os, size_t mmap_size)
+{
+	int rc = 0;
+
+	os->mmap_fd = -1;
+	os->mmap_size = 0;
+
+#ifdef OSSP_MMAP
+	if (mmap_size) {
+		char shmname[32];
+		int fd;
+		void *p;
+	
+		sprintf(shmname, "/ossp.%i", getpid());
+		fd = shm_open(shmname, O_RDWR | O_CREAT | O_EXCL | O_TRUNC,
+			      0600);
+
+		if (fd == -1) {
+			rc = -errno;
+			warn_ose(os, rc, "failed to open shared memory");
+			return rc;
+		}
+		rc = shm_unlink(shmname);
+		if (rc == -1) {
+			rc = -errno;
+			close(fd);
+			warn_ose(os, rc, "failed to unlink shared memory");
+			return rc;
+		}
+		rc = ftruncate(fd, mmap_size + 2*sizeof(struct ossp_transfer));
+		if (rc == -1) {
+			rc = -errno;
+			close(fd);
+			warn_ose(os, rc, "failed to set shared memory size");
+			return rc;
+		}
+		rc = fcntl(fd, F_SETFD, 0); /* reset cloexec */
+		if (rc == -1) {
+			rc = -errno;
+			close(fd);
+			warn_ose(os, rc, "failed to reset close-on-exec for shared memory");
+			return rc;
+		}
+		p = mmap(NULL, mmap_size + 2 * sizeof(struct ossp_transfer),
+			 PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+		if (p == MAP_FAILED) {
+			rc = -errno;
+			close(fd);
+			warn_ose(os, rc, "failed to mmap shared memory");
+			return rc;
+		}
+		os->mmap_addr[PLAY] = p;
+		os->mmap_addr[REC] = p + mmap_size / 2;
+		os->mmap_transfer = p + mmap_size;
+
+		if (sem_init(&os->mmap_transfer[PLAY].sem, 1, 0) == -1 ||
+		    sem_init(&os->mmap_transfer[REC].sem, 1, 0) == -1) {
+			rc = -errno;
+			close(fd);
+			munmap(p, mmap_size + 2 * sizeof(struct ossp_transfer));
+			warn_ose(os, rc, "failed to init shared semaphore");
+			return rc;
+		}
+
+		os->mmap = p;
+		os->mmap_size = mmap_size;
+		os->mmap_fd = fd;
+	}
+#endif
+
+	return rc;
+}
+
 static int alloc_os(size_t stream_size, size_t mmap_size, pid_t pid, uid_t pgrp,
 		    uid_t uid, gid_t gid, int cmd_sock,
 		    const int *notify, struct fuse_session *se,
@@ -959,13 +1025,13 @@ static int alloc_os(size_t stream_size, size_t mmap_size, pid_t pid, uid_t pgrp,
 	os->pgrp = pgrp;
 	os->uid = uid;
 	os->gid = gid;
-	if (mmap_size) {
-		os->mmap_off = os->id * mmap_size;
-		os->mmap_size = mmap_size;
-	}
 	os->ucnt = ucnt;
 	os->se = se;
 
+	rc = os_create_shared_memory(os, mmap_size);
+	if (rc)
+		goto err_unlock;
+
 	memset(os->vol, -1, sizeof(os->vol));
 	memset(os->vol_set, -1, sizeof(os->vol));
 
@@ -1055,56 +1121,42 @@ static void put_os(struct ossp_stream *os)
 	pthread_mutex_unlock(&mutex);
 }
 
-static void set_extra_env(pid_t pid)
+static void set_extra_env(const pid_t pid)
 {
-	char procenviron[32];
-	const int step = 1024;
-	char *data = malloc(step + 1);
-	int ofs = 0;
-	int fd;
-	int ret;
+	const char *target_vars[] = {
+		"DISPLAY=",        // Display manager
+		"PULSE_",          // PulseAudio
+		"XDG_RUNTIME_DIR=" // Audio servers
+	};
 
-	if (!data)
+	char *line = NULL;
+	FILE *file;
+	size_t size;
+	ssize_t len;
+
+	char path[32];
+	snprintf(path, sizeof(path), "/proc/%d/environ", pid);
+
+	file = fopen(path, "r");
+	if (!file)
 		return;
 
-	sprintf(procenviron, "/proc/%d/environ", pid);
-	fd = open(procenviron, O_RDONLY);
-	if (fd < 0)
-		return;
+	while ((len = getdelim(&line, &size, '\0', file)) != -1) {
+		for (uint8_t i = 0; i < ARRAY_SIZE(target_vars); ++i) {
+			char *sign;
 
-	/*
-	 * There should really be a 'read whole file to a newly allocated
-	 * buffer' function.
-	 */
-	while ((ret = read(fd, data + ofs, step)) > 0) {
-		char *newdata;
-		ofs += ret;
-		newdata = realloc(data, ofs + step + 1);
-		if (!newdata) {
-			ret = -1;
-			break;
-		}
-		data = newdata;
-	}
-	if (ret == 0) {
-		char *ptr = data;
-		/* Append the extra 0 for end condition */
-		data[ofs] = 0;
+			if (strncmp(line, target_vars[i], strlen(target_vars[i])) != 0)
+				continue;
 
-		while ((ret = strlen(ptr)) > 0) {
-			/*
-			 * Copy all PULSE variables and DISPLAY so that
-			 * ssh -X remotehost 'mplayer -ao oss' will work
-			 */
-			if (!strncmp(ptr, "DISPLAY=", 8) ||
-			    !strncmp(ptr, "PULSE_", 6))
-				putenv(ptr);
-			ptr += ret + 1;
+			if ((sign = strchr(line, '='))) {
+				*sign = '\0';
+				setenv(line, sign + 1, 1);
+			}
 		}
 	}
 
-	free(data);
-	close(fd);
+	free(line);
+	fclose(file);
 }
 
 static int create_os(const char *slave_path,
@@ -1189,12 +1241,12 @@ static int create_os(const char *slave_path,
 	if (os->slave_pid == 0) {
 		/* child */
 		char id_str[2][16], fd_str[3][16];
-		char mmap_off_str[32], mmap_size_str[32];
+		char mmap_size_str[32];
 		char log_str[16], slave_path_copy[PATH_MAX];
 		char *argv[] = { slave_path_copy, "-u", id_str[0],
 				 "-g", id_str[1], "-c", fd_str[0],
 				 "-n", fd_str[1], "-m", fd_str[2],
-				 "-o", mmap_off_str, "-s", mmap_size_str,
+				 "-s", mmap_size_str,
 				 "-l", log_str, NULL, NULL };
 		struct passwd *pwd;
 
@@ -1202,11 +1254,6 @@ static int create_os(const char *slave_path,
 		if (close(cmd_sock[0]) || close(notify_sock[0]))
 			fatal_e(-errno, "failed to close server pipe fds");
 
-#ifdef OSSP_MMAP
-		if (!mmap_size)
-			close(fuse_mmap_fd(se));
-#endif
-
 		clearenv();
 		pwd = getpwuid(os->uid);
 		if (pwd) {
@@ -1230,15 +1277,9 @@ static int create_os(const char *slave_path,
 		snprintf(id_str[1], sizeof(id_str[0]), "%d", os->gid);
 		snprintf(fd_str[0], sizeof(fd_str[0]), "%d", cmd_sock[1]);
 		snprintf(fd_str[1], sizeof(fd_str[1]), "%d", notify_sock[1]);
-		snprintf(fd_str[2], sizeof(fd_str[2]), "%d",
-#ifdef OSSP_MMAP
-			 mmap_size ? fuse_mmap_fd(se) :
-#endif
-			 -1);
-		snprintf(mmap_off_str, sizeof(mmap_off_str), "0x%llx",
-			 (unsigned long long)os->mmap_off);
+		snprintf(fd_str[2], sizeof(fd_str[2]), "%d", os->mmap_fd);
 		snprintf(mmap_size_str, sizeof(mmap_size_str), "0x%zx",
-			 mmap_size);
+			 os->mmap_size);
 		snprintf(log_str, sizeof(log_str), "%d", ossp_log_level);
 		if (ossp_log_timestamp)
 			argv[ARRAY_SIZE(argv) - 2] = "-t";
@@ -1260,13 +1301,12 @@ static int create_os(const char *slave_path,
 	}
 
 	dbg0_os(os, "CREATE slave=%d %s", os->slave_pid, slave_path);
-	dbg0_os(os, "  client=%d cmd=%d:%d notify=%d:%d mmap=%d:0x%llx:%zu",
+	dbg0_os(os, "  client=%d cmd=%d:%d notify=%d:%d mmap=%d:%zu",
 		pid, cmd_sock[0], cmd_sock[1], notify_sock[0], notify_sock[1],
-#ifdef OSSP_MMAP
-		os->mmap_size ? fuse_mmap_fd(se) :
-#endif
-		-1,
-		(unsigned long long)os->mmap_off, os->mmap_size);
+		os->mmap_fd, os->mmap_size);
+
+	close(os->mmap_fd);
+	os->mmap_fd = -1;
 
 	*osp = os;
 	rc = 0;
@@ -1510,7 +1550,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 	case OSS_GETVERSION:
 		i = SNDRV_OSS_VERSION;
 		PREP_UARG(NULL, &i);
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SNDCTL_DSP_GETCAPS:
 		i = DSP_CAP_DUPLEX | DSP_CAP_REALTIME | DSP_CAP_TRIGGER |
@@ -1519,12 +1560,14 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 #endif
 			DSP_CAP_MULTI;
 		PREP_UARG(NULL, &i);
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SNDCTL_DSP_NONBLOCK:
 		dsps->nonblock = 1;
 		ret = 0;
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	case SNDCTL_DSP_RESET:		op = OSSP_DSP_RESET;		goto nd;
 	case SNDCTL_DSP_SYNC:		op = OSSP_DSP_SYNC;		goto nd;
@@ -1533,7 +1576,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		ret = exec_simple_cmd(&dsps->os, op, NULL, NULL);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	case SOUND_PCM_READ_RATE:	op = OSSP_DSP_GET_RATE;		goto ri;
 	case SOUND_PCM_READ_BITS:	op = OSSP_DSP_GET_FORMAT;	goto ri;
@@ -1546,7 +1590,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		ret = exec_simple_cmd(&dsps->os, op, NULL, &i);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SNDCTL_DSP_SPEED:		op = OSSP_DSP_SET_RATE;		goto wi;
 	case SNDCTL_DSP_SETFMT:		op = OSSP_DSP_SET_FORMAT;	goto wi;
@@ -1557,7 +1602,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		ret = exec_simple_cmd(&dsps->os, op, &i, &i);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SNDCTL_DSP_STEREO:
 		PREP_UARG(NULL, &i);
@@ -1566,7 +1612,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		i--;
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, &i);
+		fuse_reply_ioctl(req, 0, &i, sizeof(i));
+		break;
 
 	case SNDCTL_DSP_SETFRAGMENT:
 		PREP_UARG(&i, NULL);
@@ -1574,7 +1621,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 				      OSSP_DSP_SET_FRAGMENT, &i, NULL);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	case SNDCTL_DSP_SETTRIGGER:
 		PREP_UARG(&i, NULL);
@@ -1582,7 +1630,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 				      OSSP_DSP_SET_TRIGGER, &i, NULL);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	case SNDCTL_DSP_GETOSPACE:
 	case SNDCTL_DSP_GETISPACE: {
@@ -1603,7 +1652,8 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		ret = exec_simple_cmd(&dsps->os, op, NULL, &info);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, &info);
+		fuse_reply_ioctl(req, 0, &info, sizeof(info));
+		break;
 	}
 
 	case SNDCTL_DSP_GETOPTR:
@@ -1616,14 +1666,16 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 		ret = exec_simple_cmd(&dsps->os, op, NULL, &info);
 		if (ret)
 			goto err;
-		IOCTL_RETURN(0, &info);
+		fuse_reply_ioctl(req, 0, &info, sizeof(info));
+		break;
 	}
 
 	case SNDCTL_DSP_GETODELAY:
 		PREP_UARG(NULL, &i);
 		i = 0;
 		ret = exec_simple_cmd(&dsps->os, OSSP_DSP_GET_ODELAY, NULL, &i);
-		IOCTL_RETURN(ret, &i);	/* always copy out result, 0 on err */
+		fuse_reply_ioctl(req, ret, &i, sizeof(i));	/* always copy out result, 0 on err */
+		break;
 
 	case SOUND_PCM_WRITE_FILTER:
 	case SOUND_PCM_READ_FILTER:
@@ -1638,14 +1690,16 @@ static void dsp_ioctl(fuse_req_t req, int signed_cmd, void *uarg,
 	case SNDCTL_DSP_SETSYNCRO:
 	case SNDCTL_DSP_SETDUPLEX:
 	case SNDCTL_DSP_PROFILE:
-		IOCTL_RETURN(0, NULL);
+		fuse_reply_ioctl(req, 0, NULL, 0);
+		break;
 
 	default:
 		warn_os(os, "unknown ioctl 0x%x", cmd);
 		ret = -EINVAL;
 		goto err;
 	}
-	assert(0);	/* control shouldn't reach here */
+	return;
+
 err:
 	fuse_reply_err(req, -ret);
 }
@@ -1658,9 +1712,18 @@ static int dsp_mmap_dir(int prot)
 	return PLAY;
 }
 
-static void dsp_mmap(fuse_req_t req, void *addr, size_t len, int prot,
-		     int flags, off_t offset, struct fuse_file_info *fi,
-		     uint64_t mh)
+static int dsp_dir_to_mapid(struct ossp_stream *os, int dir)
+{
+	return os->id * 2 + dir + 1;
+}
+
+static int dsp_mapid_to_dir(int map_id)
+{
+	return (map_id - 1) & 1;
+}
+
+static void dsp_mmap(fuse_req_t req, uint64_t addr, size_t len, int prot,
+		     int flags, off_t offset, struct fuse_file_info *fi)
 {
 	int dir = dsp_mmap_dir(prot);
 	struct ossp_dsp_mmap_arg arg = { };
@@ -1675,7 +1738,7 @@ static void dsp_mmap(fuse_req_t req, void *addr, size_t len, int prot,
 	}
 	dsps = os_to_dsps(os);
 
-	if (!os->mmap_off || len > os->mmap_size / 2) {
+	if (len > os->mmap_size / 2) {
 		fuse_reply_err(req, EINVAL);
 		return;
 	}
@@ -1697,13 +1760,13 @@ out_unlock:
 	pthread_mutex_unlock(&os->mmap_mutex);
 
 	if (ret == 0)
-		fuse_reply_mmap(req, os->mmap_off + dir * os->mmap_size / 2, 0);
+		fuse_reply_mmap(req, dsp_dir_to_mapid(os, dir), os->mmap_size / 2);
 	else
 		fuse_reply_err(req, -ret);
 }
 
-static void dsp_munmap(fuse_req_t req, size_t len, struct fuse_file_info *fi,
-		       off_t offset, uint64_t mh)
+static void dsp_munmap(fuse_req_t req, uint64_t map_id, size_t length,
+		       struct fuse_file_info *fi)
 {
 	struct ossp_stream *os;
 	struct ossp_dsp_stream *dsps;
@@ -1716,13 +1779,11 @@ static void dsp_munmap(fuse_req_t req, size_t len, struct fuse_file_info *fi,
 
 	pthread_mutex_lock(&os->mmap_mutex);
 
-	for (dir = 0; dir < 2; dir++)
-		if (offset == os->mmap_off + dir * os->mmap_size / 2)
-			break;
-	if (dir == 2 || len > os->mmap_size / 2) {
-		warn_os(os, "invalid munmap request "
-			"offset=%llu len=%zu mmapped=0x%x",
-			(unsigned long long)offset, len, dsps->mmapped);
+	dir = dsp_mapid_to_dir(map_id);
+	if (dsp_dir_to_mapid(os, dir) != map_id ||
+	    (dir != PLAY && dir != REC) || length != os->mmap_size / 2) {
+		warn_os(os, "invalid munmap request map_id=%llu len=%zu mmapped=0x%x",
+			(unsigned long long) map_id, length, dsps->mmapped);
 		goto out_unlock;
 	}
 
@@ -1735,7 +1796,80 @@ static void dsp_munmap(fuse_req_t req, size_t len, struct fuse_file_info *fi,
 out_unlock:
 	pthread_mutex_unlock(&os->mmap_mutex);
 out:
-	fuse_reply_none(req);
+	fuse_reply_err(req, -rc);
+}
+
+static void fill_mmap_buffer(struct ossp_stream *os)
+{
+	int rc;
+	struct fuse_chan *ch = fuse_session_next_chan(os->se, NULL);
+	unsigned mapid = dsp_dir_to_mapid(os, PLAY);
+	size_t pos = os->mmap_transfer[PLAY].pos;
+	size_t bytes = os->mmap_transfer[PLAY].bytes;
+
+	dbg1_os(os, "fill mmap buffer pos=%zu bytes=%zu", pos, bytes);
+
+	rc = fuse_lowlevel_notify_retrieve(ch, mapid, bytes, pos, os);
+	if (rc)
+		sem_post(&os->mmap_transfer[PLAY].sem);
+}
+
+static void dsp_retrieve_reply(fuse_req_t req, void *cookie, uint64_t map_id,
+			       off_t offset, struct fuse_bufvec *bufv)
+{
+	struct ossp_stream *os = cookie;
+	size_t size = fuse_buf_size(bufv);
+	struct fuse_bufvec dest = FUSE_BUFVEC_INIT(size);
+	ssize_t res;
+
+	if (dsp_mapid_to_dir(map_id) != PLAY ||
+	    dsp_dir_to_mapid(os, PLAY) != map_id ||
+	    offset != os->mmap_transfer[PLAY].pos ||
+	    size > os->mmap_transfer[PLAY].bytes) {
+		warn_os(os, "invalid retrieve reply map_id=%llu offset=%llu, size=%zu",
+			(unsigned long long) map_id,
+			(unsigned long long) offset, size);
+		sem_post(&os->mmap_transfer[PLAY].sem);	
+		return;
+	}
+
+	dest.buf[0].mem = os->mmap_addr[PLAY] + offset;
+
+	res = fuse_buf_copy(&dest, bufv, 0);
+	if (res < 0)
+		warn_ose(os, res, "dsp_retrieve_reply: buffer copy");
+	else if ((size_t) res < size) {
+		warn_os(os, "dsp_retrieve_reply: short buffer copy");
+	}
+
+	if (size < os->mmap_transfer[PLAY].bytes) {
+		os->mmap_transfer[PLAY].bytes -= size;
+		os->mmap_transfer[PLAY].pos += size;
+
+		fill_mmap_buffer(os);
+	} else {
+		sem_post(&os->mmap_transfer[PLAY].sem);	
+	}
+}
+
+static void store_mmap_buffer(struct ossp_stream *os)
+{
+	int rc;
+	struct fuse_chan *ch = fuse_session_next_chan(os->se, NULL);
+	unsigned mapid = dsp_dir_to_mapid(os, REC);
+	size_t pos = os->mmap_transfer[REC].pos;
+	size_t bytes = os->mmap_transfer[REC].bytes;
+	struct fuse_bufvec bufv = FUSE_BUFVEC_INIT(bytes);
+
+	dbg1_os(os, "store mmap buffer pos=%zu bytes=%zu", pos, bytes);
+
+	bufv.buf[0].mem = os->mmap_addr[REC] + pos;
+		
+	rc = fuse_lowlevel_notify_store(ch, mapid, pos, &bufv, 0);
+	if (rc < 0)
+		warn_ose(os, rc, "store_mmap_buffer: failure");
+
+	sem_post(&os->mmap_transfer[REC].sem);
 }
 #endif
 
@@ -1802,6 +1936,14 @@ repeat:
 				os->mixer->modify_counter++;
 				pthread_mutex_unlock(&mixer_mutex);
 				break;
+#ifdef OSSP_MMAP
+			case OSSP_NOTIFY_FILL:
+				fill_mmap_buffer(os);
+				break;
+			case OSSP_NOTIFY_STORE:
+				store_mmap_buffer(os);
+				break;
+#endif
 			default:
 			unknown:
 				warn_os(os, "unknown notification %d",
@@ -1839,6 +1981,9 @@ repeat:
 		pthread_mutex_unlock(&mutex);
 	}
 	goto repeat;
+
+	/* not reachable */
+	return NULL;
 }
 
 
@@ -1878,9 +2023,20 @@ repeat:
 
 	dbg1_os(os, "slave %d reaped", os->slave_pid);
 	__clear_bit(os->id, os_id_bitmap);
+	if (os->mmap_size) {
+		sem_destroy(&os->mmap_transfer[PLAY].sem);
+		sem_destroy(&os->mmap_transfer[REC].sem);
+
+		munmap(os->mmap, os->mmap_size + 2 * sizeof(struct ossp_transfer));
+		close(os->mmap_fd);
+	}
+		
 	free(os);
 
 	goto repeat;
+
+	/* not reachable */
+	return NULL;
 }
 
 
@@ -1972,6 +2128,7 @@ static const struct cuse_lowlevel_ops dsp_ops = {
 #ifdef OSSP_MMAP
 	.mmap			= dsp_mmap,
 	.munmap			= dsp_munmap,
+	.retrieve_reply		= dsp_retrieve_reply,
 #endif
 };
 
@@ -1985,6 +2142,7 @@ static const struct cuse_lowlevel_ops adsp_ops = {
 #ifdef OSSP_MMAP
 	.mmap			= dsp_mmap,
 	.munmap			= dsp_munmap,
+	.retrieve_reply		= dsp_retrieve_reply,
 #endif
 };
 
@@ -2068,21 +2226,18 @@ static struct fuse_session *setup_ossp_cuse(const struct cuse_lowlevel_ops *ops,
 				.flags = CUSE_UNRESTRICTED_IOCTL };
 	struct fuse_session *se;
 	int fd;
+	int multithreaded;
 
 	snprintf(name_buf, sizeof(name_buf), "DEVNAME=%s", name);
 
-	se = cuse_lowlevel_setup(argc, argv, &ci, ops, NULL, NULL);
+	se = cuse_lowlevel_setup(argc, argv, &ci, ops, &multithreaded, NULL);
 	if (!se) {
 		err("failed to setup %s CUSE", name);
 		return NULL;
 	}
 
-	fd = fuse_chan_fd(fuse_session_next_chan(se, NULL));
-	if (
-#ifdef OSSP_MMAP
-		fd != fuse_mmap_fd(se) &&
-#endif
-		fcntl(fd, F_SETFD, FD_CLOEXEC) < 0) {
+	fd = fuse_session_fd(se);
+	if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0) {
 		err_e(-errno, "failed to set CLOEXEC on %s CUSE fd", name);
 		cuse_lowlevel_teardown(se);
 		return NULL;
@@ -2096,7 +2251,7 @@ static void *cuse_worker(void *arg)
 	struct fuse_session *se = arg;
 	int rc;
 
-	rc = fuse_session_loop_mt(se);
+	rc = fuse_session_loop_mt(se, NULL);
 	cuse_lowlevel_teardown(se);
 
 	return (void *)(unsigned long)rc;
@@ -2109,7 +2264,7 @@ static int process_arg(void *data, const char *arg, int key,
 
 	switch (key) {
 	case 0:
-		fprintf(stderr, usage);
+		fputs(usage, stderr);
 		param->help = 1;
 		return 0;
 	case 1:
@@ -2166,8 +2321,7 @@ int main(int argc, char **argv)
 	/* daemonization already handled, prevent forking inside FUSE */
 	fuse_opt_add_arg(&args, "-f");
 
-	info("OSS Proxy v%s (C) 2008-2010 by Tejun Heo <teheo@suse.de>",
-	     OSSP_VERSION);
+	info("OSS Proxy v%s", OSSP_VERSION);
 
 	/* ignore stupid SIGPIPEs */
 	memset(&sa, 0, sizeof(sa));
@@ -2253,7 +2407,7 @@ int main(int argc, char **argv)
 					   param.mixer_major, param.mixer_minor,
 					   args.argc, args.argv);
 	if (strlen(param.adsp_name))
-		adsp_se = setup_ossp_cuse(&dsp_ops, param.adsp_name,
+		adsp_se = setup_ossp_cuse(&adsp_ops, param.adsp_name,
 					  param.adsp_major, param.adsp_minor,
 					  args.argc, args.argv);
 
